'''
Created on Jun 15, 2012

@author: andris
'''

# pylint: disable=E0611

import collections
import itertools
import datetime
import math
import re


__all__ = ['SASHeader']

import scipy.constants
#Planck constant times speed of light in eV*Angstroem units
HC = scipy.constants.codata.value('Planck constant in eV s') * \
    scipy.constants.codata.value('speed of light in vacuum') * 1e10

NEUTRON_WAVELENGTH_CONVERTOR = scipy.constants.codata.value('Planck constant') ** 2 * 0.5 / \
    (scipy.constants.codata.value('neutron mass')) / \
    scipy.constants.codata.value('electron volt-joule relationship') * 1e20# J

class SASHistory(object):
    def __init__(self, s=''):
        if s:
            history_oneliner = s.replace(';;', '<*doublesemicolon*>')
            history_list = [a.strip().replace('<*doublesemicolon*>', ';') \
                            for a in history_oneliner.split(';')]
            history = [a.split(':', 1) for a in history_list]
            self._list = [(float(a[0]), a[1].strip()) for a in history]
        else:
            self._list = []
    def linearize(self):
        history_text = [str(x[0]) + ': ' + x[1] for x in self._list]
        history_text = [a.replace(';', ';;') for a in history_text]
        return '; '.join(history_text)
    def add(self, label, at=None):
        """Add a new entry to the history.

        Inputs:
            label: history text
            at: time of the event. If None, the current time will be used.
        """
        if at is None:
            at = datetime.datetime.now()
        deltat = at - datetime.datetime.fromtimestamp(0, at.tzinfo)
        deltat_seconds = deltat.seconds + deltat.days * 24 * 3600 + deltat.microseconds * 1e-6
        self._list.append((deltat_seconds, label))
    def __str__(self):
        """Return the history in a human-readable format"""
        return '\n'.join([str(h[0]) + ': ' + h[1] for h in self._list])
    def __unicode__(self):
        return u'\n'.join([unicode(h[0]) + ': ' + h[1].encode('utf-8') for h in self._list])
    def pop(self):
        return self._list.pop()
    def __iter__(self):
        return self._list.__iter__()

class SASHeaderException(Exception):
    pass

class SASHeader(dict):
    """A class for holding measurement meta-data, such as sample-detector
    distance, photon energy, wavelength, beam position etc.

    This is a subclass of the Python dictionary class. It supports therefore
    dict-style indexing (header['BeamPosX']), but has several extensions:

    1) The main intent of this class is to offer a uniform interface for common
    header data (beam position, etc.), while keeping the original logic of
    various instruments. To accomplish that, we have key aliasing under the
    hood: the dict _key_aliases can define alias names for each key, e.g. EDF
    files contain the beam position as Center_1 and Center_2. Aliasing ensures
    that these can be accessed under the general fieldnames BeamPosX and
    BeamPosY. Either is read or modified, the underlying field gets read or
    modified.

    2) safe default values are supplied for undefined fields, e.g. default FSN
    is zero, default Title is "untitled" etc. See __missing__() for details

    3) summing and averaging of multiple headers is supported, compatibility is
    taken care of. See fields _fields_to_sum, _fields_to_average,
    _fields_to_collect and __add__, __iadd__, isequiv, summarize() methods.

    4) Loading/saving: if you want to load a header file, use the new_from_*()
    class methods. If you want to extend (update) an instance with another file,
    use the read_from_*() instance methods. Saving is supported as generalized
    B1 logfiles (text) and as attributes of a HDF5 group or dataset.

    Reader routines can be called only with a single filename (or corresponding
    data dict loaded elsewhere). New_from_*() methods have two possibilities of
    invocation:

    >>> header = new_from_xyz(filename_or_dict_object)

    or

    >>> header = new_from_xyz(fsn, fileformat=<default_name>, dirs=['.'])

    where fsn is either a scalar number or a sequence of scalar numbers (the
    so-called file sequence number) and fileformat is a C-style format string
    into which the fsn(s) is/are interpolated, e.g. 'org_%05d.header' for B1
    original header files. This argument ALWAYS has a default value. dirs is a
    list of folders to look for the file in. From the first two arguments,
    a full filename will be constructed and fed to sastool.misc.findfileindirs
    along with dirs. If fsn is a scalar, all exceptions raised by that function
    are propagated to the caller. If fsn is a sequence, exceptions are eaten
    and a sequence of only the successfully loaded headers will be returned.

    5) History: operations on the dataset can (and should) update the history.

    Examples:

    Load a B1 original header file:
    >>> h=SASHeader.new_from_B1_org('path/to/datafile/org_00015.header')

    Load an EDF file:
    >>> h=SASHeader.new_from_ESRF_ID02('path/to/datafile/sc3269_0_0015ccd')

    Supported keys (these are obligatory. Part of them have safe default
    values):
    :FSN:     File sequence number (integer)
    :FSNs:    List of file sequence numbers (defaults to an empty list)
    :Title:   Sample name
    :Dist:    Sample-to-detector distance in mm
    :DistCalibrated: Calibrated sample-to-detector distance in mm
    :Thickness:      Sample thickness (cm)
    :ThicknessError: Absolute error of sample thickness (cm)
    :Transm:         Sample transmission (e^{-mu*d})
    :TransmError:    Error of sample transmission
    :PosSample:      Sample position
    :Temperature:    Sample temperature
    :MeasTime:       Measurement time
    :FSNref1:        File sequence number for absolute intensity reference
    :Thicknessref1:  Thickness of the absolute intensity reference (cm)
    :Thicknessref1Error: Error of the thickness of the absolute intensity reference (cm)
    :Energy:         Nominal (apparent) photon energy (in case of neutrons, the
            "equivalent" photon energy h*c/lambda)
    :EnergyCalibrated: The calibrated photon energy
    :BeamPosX:       Position of the beam on the detector: row index (indexing
        starts from 0).
    :BeamPosY:       Position of the beam on the detector: column index
        (indexing starts from 0)
    :BeamsizeX:      Size of the beam at the sample, row direction (mm)
    :BeamsizeY:      Size of the beam at the sample, column direction (mm)
    :XPixel:         Pixel size of the detector, row direction (mm)
    :YPixel:         Pixel size of the detector, column direction (mm)
    :PixelSize:      Average pixel size of the detector: 0.5*(XPixel+YPixel)
    :Monitor:        Monitor counts
    :MonitorError:   Error of the monitor counts

    Some keys are treated specially:

    keys ending with ``Calibrated``
        If this key is not found, it is created with the value of the
        corresponding key without the ``Calibrated`` suffix. E.g. if
        ``EnergyCalibrated`` does not exist, it is created automatically and
        its value will be that of ``Energy``.

    keys ending with ``Error``
        if this key is not present, it is created with 0 value. Upon averaging
        multiple headers, these and the corresponding un-suffixed keys are
        treated specially. Namely, those found in SASHeader._fields_to_sum will
        be added (and the corresponding Error key will be calculated through
        Gaussian error-propagation), those found in SASHeader._fields_to_average
        will be averaged (weighted by the corresponding Error keys). I.e. the
        former are assumed to be *independent measurements of different
        quantities*, the latter are assumed to be *independent measurements of
        the same quantity*.

    """
    _plugins = []
    # the following define fields treated specially when adding one or more
    # headers together.
    # _fields_to_sum: these fields are to be added. The corresponding 'Error'
    # fields are used for error propagation. Field names can be strings or
    # compiled regular expression patterns.
    _fields_to_sum = ['MeasTime', 'Anode', re.compile('Monitor\w*')]
    # _fields_to_average: these fields will be averaged. Error propagation is
    # done.
    _fields_to_average = ['Transm', 'Temperature', 'BeamPosX', 'BeamPosY',
                          'Thickness', 'Thicknessref1']
    # _fields_to_collect: these will be collected.
    _fields_to_collect = {'FSN': 'FSNs'}
    # Testing equivalence
    _equiv_tests = {'Dist': lambda a, b: (abs(a - b) < 1),
                    'Energy': lambda a, b: (abs(a - b) < 1),
                    'Temperature': lambda a, b: (abs(a - b) < 0.5),
                    'Title': lambda a, b: (a == b),
                   }
    # dictionary of key aliases. Note that multi-level aliases are not allowed!
    # This is a 
    _key_aliases = None
    _protectedfields_to_copy = ['_protectedfields_to_copy', '_key_aliases',
                              '_fields_to_sum', '_fields_to_average',
                              '_fields_to_collect', '_equiv_tests']

    # -- Housekeeping methods: __init__, iterators, __missing__ etc. ----------
    @staticmethod
    def _set_default_kwargs_for_readers(kwargs):
        if 'dirs' not in kwargs:
            kwargs['dirs'] = None
        if 'experiment_type' not in kwargs:
            kwargs['experiment_type'] = None
        if 'error_on_not_found' not in kwargs:
            kwargs['error_on_not_found'] = True
        if 'generator' not in kwargs:
            kwargs['generator'] = False
        return kwargs

    def __new__(cls, *args, **kwargs):
        """Load one or more header structure. This function serves as a general
        entry point. It handles the calls
        ::

            >>> SASHeader(...)

        The ways of calling are:

        0) ``SASHeader()``: empty constructor

        1) ``SASHeader(<instance-of-SASHeader>)``: copy constructor

        2) ``SASHeader(<instance-of-dict>)``: casting constructor.

        3) ``SASHeader(<filename>, **kwargs)``: direct loading of a file

        4) ``SASHeader(<fileformat>, <fsn>, **kwargs)``: loading possibly more
        files.
        """
        print "SASHeader.__new__()"
        kwargs = SASHeader._set_default_kwargs_for_readers(kwargs)
        if not args:
            print "  empty __new__()"
            return super(SASHeader, cls).__new__(cls)
        elif (isinstance(args[0], SASHeader) or isinstance(args[0], dict)) or isinstance(args[0], tuple):
            print "  simple __new__()"
            return super(SASHeader, cls).__new__(cls)
        else:
            #everything from now on is handled by plug-ins.
            # Everything else is handled by IO plugins
            plugin = cls.get_IOplugin(args[0], 'READ')
            if len(args) == 2:
                print "  two args"
                if not isinstance(args[1], collections.Sequence):
                    fsns = [args[1]]
                else:
                    fsns = args[1]
                res = plugin.read_multi(args[0], fsns, **kwargs)
            elif len(args) == 1:
                print "  one arg"
                res = plugin.read(args[0], **kwargs)
            else:
                raise ValueError('Invalid number of positional arguments.')
            if isinstance(res, tuple):
                obj = super(SASHeader, cls).__new__(cls)
                obj.__init__(res)
                return obj
            else:
                gen = cls._read_multi(res)
                if len(args) == 2 and not isinstance(args[1], collections.Sequence):
                    return list(gen)[0]
                elif kwargs['generator'] :
                    return gen
                else:
                    return list(gen)
    @classmethod
    def _read_multi(cls, lis):
        print "SASHeader._read_multi()"
        for l in lis:
            print "SASHeader._read_multi() yields..."
            yield cls(l)
        print "SASHeader._read_multi() returns"
        return
    @classmethod
    def _autoguess_experiment_type(cls, file_or_dict):
        plugin = [p for p in cls._plugins if p.check_if_applies(file_or_dict)]
        if not plugin:
            raise SASHeaderException('No plugin can handle ' + str(file_or_dict))
        return plugin[0]
    
    @classmethod
    def get_valid_experiment_types(cls, mode='read'):
        """Get the available plugins which support 'read', 'write' or 'both'.
        """
        if mode.upper() == 'READ':
            return [p.name for p in cls._plugins if p.is_read_supported()]
        elif mode.upper() == 'WRITE':
            return [p.name for p in cls._plugins if p.is_write_supported()]
        elif mode.upper() == 'BOTH':
            return [p.name for p in cls._plugins if p.is_read_supported() and p.is_write_supported()]
        else:
            raise ValueError('invalid mode')

    def __init__(self, *args, **kwargs):
        """This constructor behaves identically to that of the superclass. If
        the first positional argument is a SASHeader, this copies over the
        protected parameters whose names are found in _protectedfields_to_copy.
        """
        if hasattr(self, '_was_init'):
            return
        self._was_init = True
        print "SASHeader.__init__()"
        self._key_aliases = {}
        kwargs = SASHeader._set_default_kwargs_for_readers(kwargs)
        if not args:
            print "  empty __init__()"
            super(SASHeader, self).__init__()
        elif isinstance(args[0], SASHeader):
            print "  __init__() from SASHeader"
            super(SASHeader, self).__init__(args[0])
            # copy over protected attributes
            for fn in args[0]._protectedfields_to_copy: #IGNORE:W0212
                attr = getattr(args[0], fn)
                if hasattr(attr, 'copy'):
                    # if the attribute has a copy method, use that. E.g. dicts.
                    setattr(self, fn, attr.copy())
                elif isinstance(attr, collections.Sequence):
                    # if the attribute is a sequence, use the [:] construct.
                    setattr(self, fn, attr[:])
                else:
                    #call the constructor to copy. Note that this can raise an
                    # exception, which is forwarded to the upper level.
                    setattr(self, fn, attr.type(attr))
        elif isinstance(args[0], tuple):
            print "  __init__() from tuple"
            self.update(args[0][0])
            self._key_aliases = args[0][1]
        else:
            print "  __init__() from plugin"
            # search for a plugin to handle this
            try:
                plugin = self.get_IOplugin(args[0], 'READ', **kwargs)
            except ValueError:
                raise NotImplementedError('__init__() not supported with args[0]==' + str(args[0]))
            else:
                d, ka = plugin.read(args[0], **kwargs)
                del self._was_init
                self.__init__((d, ka))
    def copy(self, *args, **kwargs):
        """Make a copy of this header structure"""
        d = super(SASHeader, self).copy(*args, **kwargs)
        return SASHeader(d)
    def __missing__(self, key, dry_run=False):
        """Create default values for missing fields"""
        if key in ['FSNs']:
            val = []
        elif key.endswith('Error'):
            val = 0
        elif key.startswith('Monitor'):
            val = 0
        elif key == '__particle__':
            val = 'photon'
        elif key == 'Energy':
            if not self.__contains__('Wavelength', False):
                raise KeyError(key)
            if self['__particle__'] == 'photon':
                val = HC / self['Wavelength']
            elif self['__particle__'] == 'neutron':
                val = NEUTRON_WAVELENGTH_CONVERTOR / self['Wavelength'] ** 2
            else:
                raise ValueError('Invalid particle type: ' + self['__particle__'])
        elif key == 'Wavelength':
            if not self.__contains__('Energy', False):
                raise KeyError(key)
            if self['__particle__'] == 'photon':
                val = HC / self['Energy']
            elif self['__particle__'] == 'neutron':
                val = (NEUTRON_WAVELENGTH_CONVERTOR / self['Energy']) ** 0.5
            else:
                raise ValueError('Invalid particle type: ' + self['__particle__'])
        elif key in ['maskid']:
            val = None
        elif key.startswith('FSN'):
            val = 0
        elif key == 'Title':
            val = '<untitled>'
        elif key.endswith('Calibrated'):
            val = self[key[:-len('Calibrated')]]
        #elif key in ['Dist', 'Energy', 'BeamPosX', 'BeamPosY', 'PixelSize']:
        #    val = np.NAN
        elif key in ['XPixel', 'YPixel']:
            val = self['PixelSize']
        else:
            raise KeyError(key)
        if not dry_run:
            super(SASHeader, self).__setitem__(key, val)
        return val
    def __unicode__(self):
        """Print a short summary of this header"""
        return "FSN %s; %s; %s mm; %s eV" % (self['FSN'], self['Title'], self['Dist'], self['Energy'])
    __str__ = __unicode__
    def __repr__(self):
        return "<SASHeader: " + unicode(self) + '>'
    def __getitem__(self, key):
        """ respond to header[key] requests, implements key aliasing."""
        if key in self._key_aliases:
            return super(SASHeader, self).__getitem__(self._key_aliases[key])
        else:
            return super(SASHeader, self).__getitem__(key)
    def __setitem__(self, key, value):
        """ respond to header[key]=value requests, implements key aliasing."""
        if key in self._key_aliases:
            return self.__setitem__(self._key_aliases[key], value)
        else:
            return super(SASHeader, self).__setitem__(key, value)
    def __delitem__(self, key):
        """ respond to del header[key] requests, implements key aliasing."""
        if key in self:
            return super(SASHeader, self).__delitem__(key)
        elif key in self._key_aliases:
            return self.__delitem__(self._key_aliases[key])
        else:
            raise KeyError(key)
    def __contains__(self, key, generate_missing=True):
        """ respond to 'key' in header requests, implements key aliasing."""
        if key in self._key_aliases:
            return super(SASHeader, self).__contains__(self._key_aliases[key])
        else:
            ret = super(SASHeader, self).__contains__(key)
            if not ret and generate_missing: # try if the key can be auto-generated by __missing__()
                try:
                    self.__missing__(key, dry_run=True)
                except KeyError:
                    return False
                return True
            else:
                return ret
    def __iter__(self):
        """ Return an iterator. This is used e.g. in for k in header constructs.
        """
        return self.iterkeys()
    def keys(self):
        """ Return the keys present. Alias names are also listed."""
        return [k for k in self.iterkeys()]
    def values(self):
        """ Return values. Aliases are listed more than one times."""
        return [v for v in self.itervalues()]
    def items(self):
        """ Return (key,value) pairs. Aliases are listed more than one times.
        """
        return [i for i in self.iteritems()]
    def iterkeys(self):
        """ Iterator version of keys()."""
        return itertools.chain(super(SASHeader, self).iterkeys(), self._key_aliases.iterkeys())
    def itervalues(self):
        """ Iterator version of values()."""
        return itertools.chain(super(SASHeader, self).itervalues(),
                 itertools.imap(lambda x:self[self._key_aliases[x]], self._key_aliases.iterkeys()))
    def iteritems(self):
        """ Iterator version of items()."""
        return itertools.izip(self.iterkeys(), self.itervalues())

    @classmethod
    def register_IOplugin(cls, plugin, idx=None):
        if idx is None:
            cls._plugins.append(plugin)
        else:
            cls._plugins.insert(idx, plugin)
    @classmethod
    def get_IOplugin(cls, filename, mode='READ', **kwargs):
        plugin = []
        if mode.upper() == 'READ':
            checkmode = lambda a:a.is_read_supported()
        elif mode.upper() == 'WRITE':
            checkmode = lambda a:a.is_write_supported()
        elif mode.upper() == 'BOTH':
            checkmode = lambda a:(a.is_read_supported() and a.is_write_supported())
        else:
            raise ValueError('Invalid mode!')
        if 'experiment_type' in kwargs:
            plugin = [p for p in cls._plugins if p.name == kwargs['experiment_type'] and checkmode(p)]
        if not plugin:
            plugin = [p for p in cls._plugins if p.check_if_applies(filename) and checkmode(p)]
        if not plugin:
            raise ValueError('No plugin can handle ' + str(filename))
        return plugin[0]
    def write(self, writeto, **kwargs):
        plugin = self.get_IOplugin(writeto, 'WRITE')
        plugin.write(writeto, self, **kwargs)
        
    # ------------------------ History manipulation ---------------------------

    def add_history(self, text, time=None):
        """Add a new entry to the history.

        Inputs:
            text: history text
            time: time of the event. If None, the current time will be used.
        """
        if 'History' not in self:
            self['History'] = SASHistory()
        self['History'].add(text, time)

    def get_history(self):
        """Return the history in a human-readable format"""
        return str(self['History'])


# --------------------- Summarizing, averaging and equivalence -------------

    def __iadd__(self, other):
        """Add in-place. The actual work is done by the SASHeader.summarize()
        classmethod."""
        obj = SASHeader.summarize(self, other)
        for k in obj.keys():
            self[k] = obj[k]
        return self

    def __add__(self, other):
        """Add two headers. The actual work is done by the SASHeader.summarize()
        classmethod."""
        return SASHeader.summarize(self, other)

    @classmethod
    def summarize(cls, *args):
        """Summarize several headers. Calling convention:

        summed=SASHeader.summarize(header1,header2,header3,...)

        Several fields are treated specially. Values of fields in
        SASHeader._fields_to_sum are summed (error is calculated by taking the
        corresponding 'Error' fields into account). Values of fields in
        SASHeader._fields_to_average get averaged. Values of fields in
        SASHeader._fields_to_collect get collected to a corresponding list field.
        """
        if any([not isinstance(a, SASHeader) for a in args]):
            raise NotImplementedError('Only SASHeaders can be averaged!')
        obj = cls()
        fields_treated = []
        allfieldnames = set(reduce(lambda (a, b):a + b, [a.keys() for a in args], []))
        for k in allfieldnames.intersection(cls._fields_to_sum):
            dk = k + 'Error'
            obj[k] = sum([a[k] for a in args])
            obj[dk] = math.sqrt(sum([a[dk] ** 2 for a in args]))
            fields_treated.extend([k, dk])
        allfieldnames = allfieldnames.difference(fields_treated)
        for k in allfieldnames.intersection(cls._fields_to_average):
            dk = k + 'Error'
            sumweight = sum([1 / a[dk] ** 2 for a in args])
            obj[k] = sum([a[k] / a[dk] ** 2 for a in args]) / sumweight
            obj[dk] = 1 / math.sqrt(sumweight)
            fields_treated.extend([k, dk])
        allfieldnames = allfieldnames.difference(fields_treated)
        for k in allfieldnames.intersection(cls._fields_to_collect.keys()):
            k_c = cls._fields_to_collect[k]
            obj[k_c] = sum([a[k_c] for a in args])
            obj[k_c].extend([a[k] for a in args])
            obj[k_c] = list(set(obj[k_c]))
            obj[k] = obj[k_c][0]  # take the first one.
            fields_treated.extend([k, k_c])
        allfieldnames = allfieldnames.difference(fields_treated)
        for k in allfieldnames:
            # find the first occurrence of the field
            obj[k] = [a for a in args if k in a.keys()][0][k]
        obj.add_history('Summed from: ' + ' and '.join([unicode(a) for a in args]))
        return obj

    def isequiv(self, other):
        """Test if the two headers are equivalent. The information found in
        SASHeader._equiv_tests is used to decide equivalence.
        """
        return all([self._equiv_tests[k](self[k], other[k]) for k in self._equiv_tests] + 
                   [other._equiv_tests[k](self[k], other[k]) for k in other._equiv_tests]) #IGNORE:W0212

    # ---------------------------- HDF5 I/O ----------------------------------


